@@ -3,30 +3,15 @@
  *
  * Author: Jorn Baayen <jorn@openedhand.com>
  *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Library General Public
- * License as published by the Free Software Foundation; either
- * version 2 of the License, or (at your option) any later version.
  *
- * This library is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Library General Public License for more details.
+ * SPDX-License-Identifier: LGPL-2.1-or-later
  *
- * You should have received a copy of the GNU Library General Public
- * License along with this library; if not, write to the
- * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
- * Boston, MA 02111-1307, USA.
  */
 
-/**
- * SECTION:gupnp-service-info
- * @short_description: Base abstract class for querying service information.
- *
- * The #GUPnPDeviceInfo base abstract class provides methods for querying
- * service information.
- */
 
+#define G_LOG_DOMAIN "gupnp-service-info"
+
+#include <config.h>
 #include <libsoup/soup.h>
 #include <string.h>
 
@@ -39,10 +24,6 @@
 #include "http-headers.h"
 #include "xml-util.h"
 
-G_DEFINE_ABSTRACT_TYPE (GUPnPServiceInfo,
-                        gupnp_service_info,
-                        G_TYPE_OBJECT);
-
 struct _GUPnPServiceInfoPrivate {
         GUPnPContext *context;
 
@@ -50,16 +31,29 @@ struct _GUPnPServiceInfoPrivate {
         char *udn;
         char *service_type;
 
-        SoupURI *url_base;
+        GUri *url_base;
 
         GUPnPXMLDoc *doc;
 
         xmlNode *element;
 
-        /* For async downloads */
-        GList *pending_gets;
+        GCancellable *pending_downloads_cancellable;
 };
 
+typedef struct _GUPnPServiceInfoPrivate GUPnPServiceInfoPrivate;
+
+
+/**
+ * GUPnPServiceInfo:
+ *
+ * Service information shared by local and remote services.
+ *
+ * A class that contains the common parts between local and remote services.
+ */
+G_DEFINE_ABSTRACT_TYPE_WITH_PRIVATE (GUPnPServiceInfo,
+                                     gupnp_service_info,
+                                     G_TYPE_OBJECT)
+
 enum {
         PROP_0,
         PROP_CONTEXT,
@@ -71,27 +65,13 @@ enum {
         PROP_ELEMENT
 };
 
-typedef struct {
-        GUPnPServiceInfo                 *info;
-
-        GUPnPServiceIntrospectionCallback callback;
-        gpointer                          user_data;
-
-        SoupMessage                      *message;
-} GetSCPDURLData;
-
-static void
-get_scpd_url_data_free (GetSCPDURLData *data)
-{
-        g_slice_free (GetSCPDURLData, data);
-}
-
 static void
 gupnp_service_info_init (GUPnPServiceInfo *info)
 {
-        info->priv = G_TYPE_INSTANCE_GET_PRIVATE (info,
-                                                  GUPNP_TYPE_SERVICE_INFO,
-                                                  GUPnPServiceInfoPrivate);
+        GUPnPServiceInfoPrivate *priv;
+        priv = gupnp_service_info_get_instance_private (info);
+
+        priv->pending_downloads_cancellable = g_cancellable_new ();
 }
 
 static void
@@ -101,34 +81,32 @@ gupnp_service_info_set_property (GObject      *object,
                                  GParamSpec   *pspec)
 {
         GUPnPServiceInfo *info;
+        GUPnPServiceInfoPrivate *priv;
 
         info = GUPNP_SERVICE_INFO (object);
+        priv = gupnp_service_info_get_instance_private (info);
 
         switch (property_id) {
         case PROP_CONTEXT:
-                info->priv->context = g_object_ref (g_value_get_object (value));
+                priv->context = g_object_ref (g_value_get_object (value));
                 break;
         case PROP_LOCATION:
-                info->priv->location = g_value_dup_string (value);
+                priv->location = g_value_dup_string (value);
                 break;
         case PROP_UDN:
-                info->priv->udn = g_value_dup_string (value);
+                priv->udn = g_value_dup_string (value);
                 break;
         case PROP_SERVICE_TYPE:
-                info->priv->service_type = g_value_dup_string (value);
+                priv->service_type = g_value_dup_string (value);
                 break;
         case PROP_URL_BASE:
-                info->priv->url_base = g_value_get_pointer (value);
-                if (info->priv->url_base)
-                        info->priv->url_base =
-                                soup_uri_copy (info->priv->url_base);
-
+                priv->url_base = g_value_dup_boxed (value);
                 break;
         case PROP_DOCUMENT:
-                info->priv->doc = g_value_dup_object (value);
+                priv->doc = g_value_dup_object (value);
                 break;
         case PROP_ELEMENT:
-                info->priv->element = g_value_get_pointer (value);
+                priv->element = g_value_get_pointer (value);
                 break;
         default:
                 G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
@@ -143,29 +121,27 @@ gupnp_service_info_get_property (GObject    *object,
                                  GParamSpec *pspec)
 {
         GUPnPServiceInfo *info;
+        GUPnPServiceInfoPrivate *priv;
 
         info = GUPNP_SERVICE_INFO (object);
+        priv = gupnp_service_info_get_instance_private (info);
 
         switch (property_id) {
         case PROP_CONTEXT:
-                g_value_set_object (value,
-                                    info->priv->context);
+                g_value_set_object (value, priv->context);
                 break;
         case PROP_LOCATION:
-                g_value_set_string (value,
-                                    info->priv->location);
+                g_value_set_string (value, priv->location);
                 break;
         case PROP_UDN:
-                g_value_set_string (value,
-                                    info->priv->udn);
+                g_value_set_string (value, priv->udn);
                 break;
         case PROP_SERVICE_TYPE:
                 g_value_set_string (value,
                                     gupnp_service_info_get_service_type (info));
                 break;
         case PROP_URL_BASE:
-                g_value_set_pointer (value,
-                                     info->priv->url_base);
+                g_value_set_boxed (value, priv->url_base);
                 break;
         default:
                 G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
@@ -177,41 +153,18 @@ static void
 gupnp_service_info_dispose (GObject *object)
 {
         GUPnPServiceInfo *info;
+        GUPnPServiceInfoPrivate *priv;
 
         info = GUPNP_SERVICE_INFO (object);
+        priv = gupnp_service_info_get_instance_private (info);
 
         /* Cancel any pending SCPD GETs */
-        if (info->priv->context) {
-                SoupSession *session;
-
-                session = gupnp_context_get_session (info->priv->context);
-
-                while (info->priv->pending_gets) {
-                        GetSCPDURLData *data;
-
-                        data = info->priv->pending_gets->data;
-
-                        soup_session_cancel_message (session,
-                                                     data->message,
-                                                     SOUP_STATUS_CANCELLED);
-
-                        g_object_unref (data->message);
-                        get_scpd_url_data_free (data);
-
-                        info->priv->pending_gets =
-                                g_list_delete_link (info->priv->pending_gets,
-                                                    info->priv->pending_gets);
-                }
-
-                /* Unref context */
-                g_object_unref (info->priv->context);
-                info->priv->context = NULL;
+        if (!g_cancellable_is_cancelled (priv->pending_downloads_cancellable)) {
+                g_cancellable_cancel (priv->pending_downloads_cancellable);
         }
 
-        if (info->priv->doc) {
-                g_object_unref (info->priv->doc);
-                info->priv->doc = NULL;
-        }
+        g_clear_object (&priv->context);
+        g_clear_object (&priv->doc);
 
         G_OBJECT_CLASS (gupnp_service_info_parent_class)->dispose (object);
 }
@@ -220,14 +173,17 @@ static void
 gupnp_service_info_finalize (GObject *object)
 {
         GUPnPServiceInfo *info;
+        GUPnPServiceInfoPrivate *priv;
 
         info = GUPNP_SERVICE_INFO (object);
+        priv = gupnp_service_info_get_instance_private (info);
 
-        g_free (info->priv->location);
-        g_free (info->priv->udn);
-        g_free (info->priv->service_type);
+        g_clear_object (&priv->pending_downloads_cancellable);
+        g_free (priv->location);
+        g_free (priv->udn);
+        g_free (priv->service_type);
 
-        soup_uri_free (info->priv->url_base);
+        g_uri_unref (priv->url_base);
 
         G_OBJECT_CLASS (gupnp_service_info_parent_class)->finalize (object);
 }
@@ -244,8 +200,6 @@ gupnp_service_info_class_init (GUPnPServiceInfoClass *klass)
         object_class->dispose      = gupnp_service_info_dispose;
         object_class->finalize     = gupnp_service_info_finalize;
 
-        g_type_class_add_private (klass, sizeof (GUPnPServiceInfoPrivate));
-
         /**
          * GUPnPServiceInfo:context:
          *
@@ -302,7 +256,7 @@ gupnp_service_info_class_init (GUPnPServiceInfoClass *klass)
                                       G_PARAM_STATIC_BLURB));
 
         /**
-         * GUPnPServiceInfo:service-type
+         * GUPnPServiceInfo:service-type:
          *
          * The service type.
          **/
@@ -324,17 +278,17 @@ gupnp_service_info_class_init (GUPnPServiceInfoClass *klass)
          *
          * The URL base (#SoupURI).
          **/
-        g_object_class_install_property
-                (object_class,
-                 PROP_URL_BASE,
-                 g_param_spec_pointer ("url-base",
-                                       "URL base",
-                                       "The URL base",
-                                       G_PARAM_READWRITE |
-                                       G_PARAM_CONSTRUCT_ONLY |
-                                       G_PARAM_STATIC_NAME |
-                                       G_PARAM_STATIC_NICK |
-                                       G_PARAM_STATIC_BLURB));
+        g_object_class_install_property (
+                object_class,
+                PROP_URL_BASE,
+                g_param_spec_boxed ("url-base",
+                                    "URL base",
+                                    "The URL base",
+                                    G_TYPE_URI,
+                                    G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY |
+                                            G_PARAM_STATIC_NAME |
+                                            G_PARAM_STATIC_NICK |
+                                            G_PARAM_STATIC_BLURB));
 
         /**
          * GUPnPServiceInfo:document:
@@ -384,14 +338,18 @@ gupnp_service_info_class_init (GUPnPServiceInfoClass *klass)
  *
  * Get the #GUPnPContext associated with @info.
  *
- * Returns: A #GUPnPContext.
+ * Returns: (transfer none): A #GUPnPContext.
  **/
 GUPnPContext *
 gupnp_service_info_get_context (GUPnPServiceInfo *info)
 {
+        GUPnPServiceInfoPrivate *priv;
+
         g_return_val_if_fail (GUPNP_IS_SERVICE_INFO (info), NULL);
 
-        return info->priv->context;
+        priv = gupnp_service_info_get_instance_private (info);
+
+        return priv->context;
 }
 
 /**
@@ -405,9 +363,13 @@ gupnp_service_info_get_context (GUPnPServiceInfo *info)
 const char *
 gupnp_service_info_get_location (GUPnPServiceInfo *info)
 {
+        GUPnPServiceInfoPrivate *priv;
+
         g_return_val_if_fail (GUPNP_IS_SERVICE_INFO (info), NULL);
 
-        return info->priv->location;
+        priv = gupnp_service_info_get_instance_private (info);
+
+        return priv->location;
 }
 
 /**
@@ -418,12 +380,16 @@ gupnp_service_info_get_location (GUPnPServiceInfo *info)
  *
  * Returns: A constant #SoupURI.
  **/
-const SoupURI *
+const GUri *
 gupnp_service_info_get_url_base (GUPnPServiceInfo *info)
 {
+        GUPnPServiceInfoPrivate *priv;
+
         g_return_val_if_fail (GUPNP_IS_SERVICE_INFO (info), NULL);
 
-        return info->priv->url_base;
+        priv = gupnp_service_info_get_instance_private (info);
+
+        return priv->url_base;
 }
 
 /**
@@ -437,9 +403,13 @@ gupnp_service_info_get_url_base (GUPnPServiceInfo *info)
 const char *
 gupnp_service_info_get_udn (GUPnPServiceInfo *info)
 {
+        GUPnPServiceInfoPrivate *priv;
+
         g_return_val_if_fail (GUPNP_IS_SERVICE_INFO (info), NULL);
 
-        return info->priv->udn;
+        priv = gupnp_service_info_get_instance_private (info);
+
+        return priv->udn;
 }
 
 /**
@@ -448,36 +418,51 @@ gupnp_service_info_get_udn (GUPnPServiceInfo *info)
  *
  * Get the UPnP service type, or %NULL.
  *
+ * Example: `urn:schemas-upnp-org:service:RenderingControl:1`
+ *
  * Returns: A constant string.
  **/
 const char *
 gupnp_service_info_get_service_type (GUPnPServiceInfo *info)
 {
+        GUPnPServiceInfoPrivate *priv;
+
         g_return_val_if_fail (GUPNP_IS_SERVICE_INFO (info), NULL);
 
-        if (!info->priv->service_type) {
-                info->priv->service_type =
+        priv = gupnp_service_info_get_instance_private (info);
+
+        if (!priv->service_type) {
+                priv->service_type =
                         xml_util_get_child_element_content_glib
-                                (info->priv->element, "serviceType");
+                                (priv->element, "serviceType");
         }
 
-        return info->priv->service_type;
+        return priv->service_type;
 }
 
 /**
  * gupnp_service_info_get_id:
  * @info: A #GUPnPServiceInfo
  *
- * Get the ID of this service, or %NULL if there is no ID.
+ * Get the serviceID of this service, or %NULL if there is no ID.
+ *
+ * The serviceID should be unique to a device. This makes it possible to provide
+ * the same serviceType multiple times on one device
+ *
+ * Example: `org:serviceId:RenderingControl`
  *
  * Return value: A string. This string should be freed with g_free() after use.
  **/
 char *
 gupnp_service_info_get_id (GUPnPServiceInfo *info)
 {
+        GUPnPServiceInfoPrivate *priv;
+
         g_return_val_if_fail (GUPNP_IS_SERVICE_INFO (info), NULL);
 
-        return xml_util_get_child_element_content_glib (info->priv->element,
+        priv = gupnp_service_info_get_instance_private (info);
+
+        return xml_util_get_child_element_content_glib (priv->element,
                                                         "serviceId");
 }
 
@@ -492,11 +477,15 @@ gupnp_service_info_get_id (GUPnPServiceInfo *info)
 char *
 gupnp_service_info_get_scpd_url (GUPnPServiceInfo *info)
 {
+        GUPnPServiceInfoPrivate *priv;
+
         g_return_val_if_fail (GUPNP_IS_SERVICE_INFO (info), NULL);
 
-        return xml_util_get_child_element_content_url (info->priv->element,
+        priv = gupnp_service_info_get_instance_private (info);
+
+        return xml_util_get_child_element_content_url (priv->element,
                                                        "SCPDURL",
-                                                       info->priv->url_base);
+                                                       priv->url_base);
 }
 
 /**
@@ -510,11 +499,15 @@ gupnp_service_info_get_scpd_url (GUPnPServiceInfo *info)
 char *
 gupnp_service_info_get_control_url (GUPnPServiceInfo *info)
 {
+        GUPnPServiceInfoPrivate *priv;
+
         g_return_val_if_fail (GUPNP_IS_SERVICE_INFO (info), NULL);
 
-        return xml_util_get_child_element_content_url (info->priv->element,
+        priv = gupnp_service_info_get_instance_private (info);
+
+        return xml_util_get_child_element_content_url (priv->element,
                                                        "controlURL",
-                                                       info->priv->url_base);
+                                                       priv->url_base);
 }
 
 /**
@@ -528,209 +521,173 @@ gupnp_service_info_get_control_url (GUPnPServiceInfo *info)
 char *
 gupnp_service_info_get_event_subscription_url (GUPnPServiceInfo *info)
 {
+        GUPnPServiceInfoPrivate *priv;
+
         g_return_val_if_fail (GUPNP_IS_SERVICE_INFO (info), NULL);
 
-        return xml_util_get_child_element_content_url (info->priv->element,
+        priv = gupnp_service_info_get_instance_private (info);
+
+        return xml_util_get_child_element_content_url (priv->element,
                                                        "eventSubURL",
-                                                       info->priv->url_base);
+                                                       priv->url_base);
 }
 
-/**
- * gupnp_service_info_get_introspection:
- * @info: A #GUPnPServiceInfo
- * @error: return location for a #GError, or %NULL
- *
- * Note that introspection object is created from the information in service
- * description document (SCPD) provided by the service so it can not be created
- * if the service does not provide an SCPD.
- *
- * Warning: You  should use gupnp_service_info_get_introspection_async()
- * instead, this function re-enter the GMainloop before returning.
- *
- * Return value: (transfer full):  A new #GUPnPServiceIntrospection for this
- * service or %NULL. Unref after use.
- **/
-GUPnPServiceIntrospection *
-gupnp_service_info_get_introspection (GUPnPServiceInfo *info,
-                                      GError          **error)
-{
-        GUPnPServiceIntrospection *introspection;
-        SoupSession *session;
-        SoupMessage *msg;
-        int status;
-        char *scpd_url;
-        xmlDoc *scpd;
-
-        g_return_val_if_fail (GUPNP_IS_SERVICE_INFO (info), NULL);
-
-        introspection = NULL;
-
-        scpd_url = gupnp_service_info_get_scpd_url (info);
-
-        msg = NULL;
-        if (scpd_url != NULL) {
-                msg = soup_message_new (SOUP_METHOD_GET, scpd_url);
-
-                g_free (scpd_url);
-        }
-
-        if (msg == NULL) {
-                g_set_error (error,
-                             GUPNP_SERVER_ERROR,
-                             GUPNP_SERVER_ERROR_INVALID_URL,
-                             "No valid SCPD URL defined");
-
-                return NULL;
-        }
-
-        /* Send off the message */
-        session = gupnp_context_get_session (info->priv->context);
-
-        status = soup_session_send_message (session, msg);
-        if (!SOUP_STATUS_IS_SUCCESSFUL (status)) {
-                _gupnp_error_set_server_error (error, msg);
-
-                g_object_unref (msg);
-
-                return NULL;
-        }
-
-        scpd = xmlRecoverMemory (msg->response_body->data,
-                                 msg->response_body->length);
-
-        g_object_unref (msg);
-
-        if (scpd) {
-                introspection = gupnp_service_introspection_new (scpd);
-
-                xmlFreeDoc (scpd);
-        }
-
-        if (!introspection) {
-                g_set_error (error,
-                             GUPNP_SERVER_ERROR,
-                             GUPNP_SERVER_ERROR_INVALID_RESPONSE,
-                             "Could not parse SCPD");
-        }
-
-        return introspection;
-}
-
-/*
- * SCPD URL downloaded.
- */
 static void
-got_scpd_url (SoupSession    *session,
-              SoupMessage    *msg,
-              GetSCPDURLData *data)
+get_scpd_document_finished (GObject *source,
+                            GAsyncResult *res,
+                            gpointer user_data)
 {
-        GUPnPServiceIntrospection *introspection;
-        GError *error;
+        GError *error = NULL;
+        GTask *task = G_TASK (user_data);
+        xmlDoc *scpd = NULL;
 
-        introspection = NULL;
-        error = NULL;
+        GBytes *bytes =
+                soup_session_send_and_read_finish (SOUP_SESSION (source),
+                                                   res,
+                                                   &error);
 
-        if (msg->status_code == SOUP_STATUS_CANCELLED)
-                return;
+        if (error != NULL) {
+                g_task_return_error (task, error);
 
-        if (SOUP_STATUS_IS_SUCCESSFUL (msg->status_code)) {
-                xmlDoc *scpd;
+                goto out;
+        }
 
-                scpd = xmlRecoverMemory (msg->response_body->data,
-                                         msg->response_body->length);
-                if (scpd) {
-                        introspection = gupnp_service_introspection_new (scpd);
+        SoupMessage *message =
+                soup_session_get_async_result_message (SOUP_SESSION (source),
+                                                       res);
+        if (!SOUP_STATUS_IS_SUCCESSFUL (soup_message_get_status (message))) {
+                g_task_return_error (task,
+                                     _gupnp_error_new_server_error (message));
 
-                        xmlFreeDoc (scpd);
-                }
+                goto out;
+        }
 
-                if (!introspection) {
-                        error = g_error_new
-                                        (GUPNP_SERVER_ERROR,
+        gsize length;
+        gconstpointer data = g_bytes_get_data (bytes, &length);
+        scpd = xmlRecoverMemory (data, length);
+        if (scpd == NULL) {
+                g_task_return_new_error (task,
+                                         GUPNP_SERVER_ERROR,
                                          GUPNP_SERVER_ERROR_INVALID_RESPONSE,
                                          "Could not parse SCPD");
-                }
-        } else
-                error = _gupnp_error_new_server_error (msg);
 
-        data->info->priv->pending_gets =
-                g_list_remove (data->info->priv->pending_gets, data);
+                goto out;
+        }
 
-        data->callback (data->info,
-                        introspection,
-                        error,
-                        data->user_data);
+        GUPnPServiceIntrospection *introspection =
+                gupnp_service_introspection_new (scpd, &error);
 
-        if (error)
-                g_error_free (error);
+        if (error != NULL) {
+                g_task_return_error (task, error);
 
-        get_scpd_url_data_free (data);
+                goto out;
+        }
+
+        g_task_return_pointer (task, introspection, g_object_unref);
+
+out:
+        g_clear_pointer (&scpd, xmlFreeDoc);
+        g_clear_pointer (&bytes, g_bytes_unref);
+        g_object_unref (task);
 }
 
 /**
- * gupnp_service_info_get_introspection_async:
+ * gupnp_service_info_introspect_async:
  * @info: A #GUPnPServiceInfo
+ * @cancellable: (nullable) : a #GCancellable that can be used to cancel the call.
  * @callback: (scope async) : callback to be called when introspection object is ready.
  * @user_data: user_data to be passed to the callback.
  *
  * Note that introspection object is created from the information in service
  * description document (SCPD) provided by the service so it can not be created
- * if the service does not provide an SCPD.
+ * if the service does not provide a SCPD.
+ *
+ * If @cancellable is used to cancel the call, @callback will be called with
+ * error code %G_IO_ERROR_CANCELLED.
+ *
+ * Since: 1.2.2
  **/
 void
-gupnp_service_info_get_introspection_async
-                                (GUPnPServiceInfo                 *info,
-                                 GUPnPServiceIntrospectionCallback callback,
-                                 gpointer                          user_data)
+gupnp_service_info_introspect_async           (GUPnPServiceInfo    *info,
+                                               GCancellable        *cancellable,
+                                               GAsyncReadyCallback  callback,
+                                               gpointer             user_data)
 {
-        GetSCPDURLData *data;
-        char *scpd_url;
-        SoupSession *session;
+        GTask *task = g_task_new (info, cancellable, callback, user_data);
+        g_task_set_name (task, "UPnP service introspection");
 
-        g_return_if_fail (GUPNP_IS_SERVICE_INFO (info));
-        g_return_if_fail (callback != NULL);
-
-        data = g_slice_new (GetSCPDURLData);
-
-        scpd_url = gupnp_service_info_get_scpd_url (info);
-
-        data->message = NULL;
-        if (scpd_url != NULL) {
-                data->message = soup_message_new (SOUP_METHOD_GET, scpd_url);
-
-                g_free (scpd_url);
-        }
-
-        if (data->message == NULL) {
-                GError *error;
-
-                error = g_error_new
-                                (GUPNP_SERVER_ERROR,
-                                 GUPNP_SERVER_ERROR_INVALID_URL,
-                                 "No valid SCPD URL defined");
-
-                callback (info, NULL, error, user_data);
-
-                g_error_free (error);
-
-                g_slice_free (GetSCPDURLData, data);
+        char *scpd_url = gupnp_service_info_get_scpd_url (info);
+        if (scpd_url == NULL) {
+                g_task_return_new_error (task,
+                                         GUPNP_SERVER_ERROR,
+                                         GUPNP_SERVER_ERROR_INVALID_URL,
+                                         "%s",
+                                         "No valid SCPD URL defined");
+                g_object_unref (task);
 
                 return;
         }
 
-        data->info      = info;
-        data->callback  = callback;
-        data->user_data = user_data;
+        GUPnPContext *context = gupnp_service_info_get_context (info);
+        GUri *scpd = gupnp_context_rewrite_uri_to_uri (context, scpd_url);
+        g_free (scpd_url);
+
+        SoupMessage *message = soup_message_new_from_uri (SOUP_METHOD_GET, scpd);
+        g_uri_unref (scpd);
+
+        if (message == NULL) {
+                g_task_return_new_error (task,
+                                         GUPNP_SERVER_ERROR,
+                                         GUPNP_SERVER_ERROR_INVALID_URL,
+                                         "%s",
+                                         "No valid SCPD URL defined");
+                g_object_unref (task);
+
+                return;
+        }
+
+        GCancellable *internal_cancellable = g_cancellable_new ();
+        if (cancellable != NULL) {
+                g_cancellable_connect (cancellable,
+                                       G_CALLBACK (g_cancellable_cancel),
+                                       internal_cancellable,
+                                       NULL);
+        }
 
         /* Send off the message */
-        info->priv->pending_gets =
-                g_list_prepend (info->priv->pending_gets,
-                                data);
+        GUPnPServiceInfoPrivate *priv =
+                gupnp_service_info_get_instance_private (info);
+        soup_session_send_and_read_async (
+                gupnp_context_get_session (priv->context),
+                message,
+                G_PRIORITY_DEFAULT,
+                internal_cancellable,
+                get_scpd_document_finished,
+                task);
+        g_object_unref (message);
+        g_object_unref (internal_cancellable);
+}
 
-        session = gupnp_context_get_session (info->priv->context);
+/**
+ * gupnp_service_info_introspect_finish:
+ * @info: A GUPnPServiceInfo
+ * @res: A #GAsyncResult
+ * @error: (inout)(optional): Return location for a #GError, or %NULL
+ *
+ * Finish an asynchronous call initiated with
+ * gupnp_service_info_introspect_async().
+ *
+ * Returns: (nullable)(transfer full): %NULL, if the call had an error, a
+ * #GUPnPServiceIntrospection object otherwise.
+ *
+ * Since: 1.2.2
+ */
+GUPnPServiceIntrospection *
+gupnp_service_info_introspect_finish          (GUPnPServiceInfo   *info,
+                                               GAsyncResult       *res,
+                                               GError            **error)
+{
+        g_return_val_if_fail (g_task_is_valid (res, info), NULL);
 
-        soup_session_queue_message (session,
-                                    data->message,
-                                    (SoupSessionCallback) got_scpd_url,
-                                    data);
+        return g_task_propagate_pointer (G_TASK (res), error);
 }
